import re
from kevin_toolbox.computer_science.algorithm.for_nested_dict_list.name_handler import escape_node


def build_name(root_node, method_ls, node_ls, b_escape_node=True):
    """
        根据取值方式 method_ls 和取值时使用的键 node_ls，来构造名字 name
            名字 name 的具体介绍参见函数 parse_name()

        参数：
            root_node:          <str> 变量名
            method_ls:          <list of str> 取值方式，可选值为 :|@
            node_ls:            <list> 取值时使用的键
            b_escape_node:      <boolean> 是否尝试对取值的键名 node 进行转义
                                    默认为 True
        返回：
            name
    """
    assert isinstance(root_node, (str,))
    assert len(method_ls) == len(node_ls)
    if b_escape_node:
        node_ls = [escape_node(node=i, b_reversed=False, times=1) for i in node_ls]

    temp = [root_node]
    for i in zip(method_ls, node_ls):
        temp.extend(i)
    name = "".join(temp)

    return name



import copy
from kevin_toolbox.computer_science.algorithm.registration import Registry
from kevin_toolbox.computer_science.algorithm import for_nested_dict_list as fndl


class Trigger:
    """
        触发器
            通过 self.update_by_state() 来接收并监视 state_dict，
            当 state_dict 全部或部分发生变化时，将会把【发生变化的部分】
            传入 self.bind() 绑定的函数中，并执行一次该函数。
    """

    def __init__(self, **kwargs):
        """
            参数：
                target_s:           <dict> 待绑定目标
                                        以名称 name 为键，以待绑定目标 target 为值
                                        主要调用 self.bind()
                init_state          <dict> 初始状态
                                        默认不设置，此时将在第一次调用 self.update_by_state() 时设置
        """
        # 默认参数
        paras = {
            #
            "target_s": dict(),
            "init_state": dict()
        }

        # 获取参数
        paras.update(kwargs)

        self.last_state = paras["init_state"]
        self.database = Registry(uid=None)
        for k, v in paras["target_s"].items():
            self.bind(name=k, target=v)

    def bind(self, name, target):
        """
            绑定触发目标
                触发目标有以下类型
                - 可以是实例（要求实例具有 update_by_state() 方法，将该方法作为真正的触发目标）
                - 函数

            参数：
                name:               <str> 名称，具体参见 for_nested_dict_list 中关于 name 的介绍
                target:             <obj/func> 待绑定目标
        """
        func = getattr(target, "update_by_state", None)
        target = target if func is None else func
        assert callable(target), \
            f'object {target} is not callable and has no update_by_state() method to be bound'
        assert isinstance(name, (str,))

        self.database.add(obj=target, name=name, b_force=True, b_execute_now=True)

    def unbind(self, name, b_not_exist_ok=True):
        """
            接触绑定

            参数：
                name:               <str> 待接触绑定目标的名称
                b_not_exist_ok:     <boolean> name 不存在时是否不报错
                                        默认为 True，不报错
        """

        try:
            self.database.get(name=name, b_pop=True)
        except:
            if not b_not_exist_ok:
                raise IndexError(f'name {name} not exists')

    def update_by_state(self, cur_state, target_names=None):
        """
            更新状态，决定是否触发

            参数：
                cur_state:          <dict> 状态
                target_names:       <list of str> 需要更新的目标
                                        默认为 None,此时将更新所有目标
        """
        assert isinstance(cur_state, (dict,))
        target_names = [""] if target_names is None else target_names
        assert isinstance(target_names, (list,))

        new_state = dict()
        for key, value in cur_state.items():
            if key not in self.last_state or self.last_state[key] != value:
                new_state[key] = value

        if len(new_state) > 0:
            # get funcs to update
            func_s = dict()
            for target_name in target_names:
                var = self.database.get(name=target_name, b_pop=False, default=None)
                if isinstance(var, (list, dict,)):
                    func_s.update({target_name + k: v for k, v in fndl.get_nodes(var=var, level=-1, b_strict=True)})
                else:
                    func_s[target_name] = var
            # update
            for func in func_s.values():
                if func is not None:
                    func(new_state)
            self.last_state.update(new_state)

        return new_state

    # ---------------------- 用于保存和加载触发器的状态 ---------------------- #

    def load_state_dict(self, state_dict):
        """
            加载触发器状态
        """
        self.last_state = state_dict["last_state"]

    def state_dict(self):
        """
            获取触发器状态
        """
        return {"last_state": copy.deepcopy(self.last_state)}

    def clear_state_dict(self):
        self.last_state.clear()



import re
import os
import inspect
import pkgutil
import weakref
from kevin_toolbox.computer_science.algorithm import for_nested_dict_list as fndl


class Registry:
    """
        注册器
            具有以下功能
            - 管理成员，包括添加 add()、获取 get() pop() 成员等
            - 支持通过装饰器 register() 来添加成员
            - 支持通过 collect_from_paths() 搜索指定的路径，当该路径下的模块被 register() 装饰器包裹或者通过 add() 添加成员时，将自动导入
            （用于解决python中的模块是惰性的问题）

        使用方法：
            以目录结构：
                xx/——— modules_dir
                 |       |——— a.py
                 |       |___ ...
                 |——— test.py
            为例。
            我们的目标是在 test.py 中，构建一个能够自动加载 modules_dir 中待注册成员的 Registry 实例。
            具体步骤如下：
                1. 在文件 test.py 中：
                    # 创建注册器实例
                    DB = Registry(uid="DB")
                    # 设置搜索路径
                    DB.collect_from_paths(path_ls=["xx/modules_dir"])

                2. 在 modules_dir 下需要加载的成员中，以 a.py 为例：
                    # 导入注册器实例
                    from xx.test import DB
                    # 使用 DB.register() 装饰器注册
                    @DB.register(name=":my_module")
                    class A:
                        ...
                    # 使用 add() 函数注册
                    DB.add(obj=233, name=":var:c")

                3. 在文件 test.py 中：
                    # 获取已注册成员
                    module = DB.get(name=":my_module")

                4. 如果需要在其他文件中获取已有注册器实例，除了用 import 之外，还可以直接用 uid
                    # 获取指定 uid 下的实例
                    temp = Registry(uid="DB")
    """
    name = "registry made by kevin"
    __instances = weakref.WeakValueDictionary()  # {uid: instance_i, ...} 用于保存创建的实例
    __counter = 0

    def __new__(cls, *args, **kwargs):
        """
            __new__函数返回的实例，将作为self参数被传入到__init__函数。
                如果__new__函数返回一个已经存在的实例（不论是哪个类的），__init__还是会被调用的，所以要特别注意__init__中对变量的赋值。
        """
        if "uid" in kwargs:
            uid = kwargs["uid"]
        else:
            while cls.__counter in cls.__instances:
                cls.__counter += 1
            uid = cls.__counter

        # 获取实例
        if uid in cls.__instances:
            # 返回已存在的实例
            self = cls.__instances[uid]
        else:
            # 传入 __init__ 中新建一个实例
            self = super().__new__(cls)
        return self

    def __init__(self, *args, **kwargs):
        """
            参数：
                uid:                <hashable> 实例的唯一标识符
                                        - 不设置。新建一个实例，并自动分配一个整数作为 uid（一般是此时已有实例的数量-1），
                                            并将该实例记录到类变量 instances_of_Registry 中。
                                        - 设置为 None。新建一个实例，但不记录到类变量 instances_of_Registry 中。
                                        - 设置为其他值。根据 uid 到类变量 instances_of_Registry 的已有实例中寻找相同的实例，
                                            若命中则返回已有实例，若无则以该 uid 新建一个实例并添加到 instances_of_Registry 中。
                                            （使用该特性可以构造单例模式）
                                        默认为不设置。
        """
        try:
            getattr(self, "uid")
        except:
            pass
        else:
            return  # 如果是从 __new__ 中获取的已有的实例则不重复进行参数赋值
        self.database = dict()
        #
        self.uid = kwargs.get("uid", Registry.__counter)
        self._path_to_collect = []
        self._item_to_add = []
        # 记录到 __instances 中
        if self.uid is not None:
            Registry.__instances[self.uid] = self

    def add(self, obj, name=None, b_force=False, b_execute_now=True):
        """
            注册

            参数：
                obj：            待注册成员
                                    可以是函数、类或者callable的实例
                                    也可以是各种 int、float、str变量
                                    总之一切对象皆可
                name：           <str> 成员名称
                                    默认为 None，此时将从被注册对象 obj 的属性中尝试推断出其名称。
                                        若 obj 中有 name 或者 __name__ 属性（优先选择name），则推断出的名称是 f':{obj.name}'；
                                            进一步若有 version 属性，则为 f':{obj.name}:{obj.version}'
                                        否则报错。
                                        比如下面的类：
                                            class A:
                                                version="1.0"
                                        的默认注册名称将是 ":A:1.0"
                                    对于 int、str 和其他没有 name 或者 __name__ 属性的变量则必须要手动指定 name 参数。
                        需要注意的是，成员的名称确定了其在注册器内部 database 中的位置，名称的解释方式参考 get_value_by_name() 中的介绍。
                        因此不同的名称可能指向了同一个位置。
                b_force：          <boolean> 是否强制注册
                                    默认为 False，此时当 name 指向的位置上已经有成员或者需要强制修改database结构时，将不进行覆盖而直接跳过，注册失败
                                    当设置为 True，将会强制覆盖
                b_execute_now:      <boolean> 现在就执行注册
                                        默认为 True，否则将等到第一次执行 get() 函数时才会真正尝试注册

            返回：
                <boolean>   是否注册成功
        """
        # 检验参数
        if name is None:
            name = getattr(obj, "name", getattr(obj, "__name__", None))
            if name is not None:
                name = f':{name}'
                version = getattr(obj, "version", None)
                if version is not None:
                    name += f':{version}'
        assert isinstance(name, (str,))

        #
        if not b_execute_now:
            self._item_to_add.append(dict(obj=obj, name=name, b_force=b_force, b_execute_now=True))
            return True

        # 尝试注册
        temp = fndl.set_value_by_name(var=fndl.copy_(var=self.database, b_deepcopy=False), name=name, value=obj,
                                      b_force=True)
        # check
        if not b_force:
            inc_node_nums = fndl.count_leaf_node_nums(var=obj) if isinstance(obj, (list, dict)) else 1  # 应该增加的节点数量
            if fndl.count_leaf_node_nums(var=temp) != fndl.count_leaf_node_nums(var=self.database) + inc_node_nums:
                # print(f'registration failed, name {name} may be a conflict with an '
                #       f'existing member in {[i for i, j in fndl.get_nodes(var=self.database)]}')
                return False

        self.database = temp
        return True

    def get(self, name, b_pop=False, **kwargs):
        """
            获取

            参数：
                name:           <str> 成员名称
                b_pop:          <boolean> 取值的同时移除该成员
                                    默认为 False
                default:        默认值
                                    找不到时，若无默认值则报错，否则将返回默认值
        """
        # 加载待注册成员
        if len(self._item_to_add) > 0:
            for i in self._item_to_add:
                self.add(**i)
            self._item_to_add.clear()
        if len(self._path_to_collect) > 0:
            for i in self._path_to_collect:
                self.collect_from_paths(**i)
            self._path_to_collect.clear()

        return fndl.get_value_by_name(var=self.database, name=name, b_pop=b_pop, **kwargs)

    def clear(self):
        self.database.clear()

    # -------------------- 装饰器 --------------------- #

    def register(self, name=None, b_force=False, b_execute_now=True):
        """
            用于注册成员的装饰器
                成员可以是函数、类或者callable的实例

            参数：
                name：           <str> 成员名称
                                    默认为 None
                b_force：        <boolean> 是否强制注册
                                    默认为 False
                b_execute_now:  <boolean> 现在就执行注册
                                    默认为 True
                                （以上参数具体参考 add() 函数介绍）
        """

        def wrapper(obj):
            nonlocal self, name, b_force, b_execute_now
            self.add(obj, name=name, b_force=b_force, b_execute_now=b_execute_now)
            return obj

        return wrapper

    # -------------------- 通过路径添加 --------------------- #

    def collect_from_paths(self, path_ls=None, path_ls_to_exclude=None, b_execute_now=False):
        """
            遍历 path_ls 下的所有模块，并自动导入其中主要被注册的部分
                比如被 register() 装饰器包裹或者通过 add() 添加的部分

            参数：
                path_ls:            <list of paths> 需要搜索的目录
                path_ls_to_exclude: <list of paths> 需要排除的目录
                b_execute_now:      <boolean> 现在就执行导入
                                        默认为 False，将等到第一次执行 get() 函数时才会真正尝试导入

                        注意，在某个文件中使用 Registry 实例时，应尽量避免下面的情况：
                            1. 在当前脚本中显式导入该实例前，调用了其他脚本执行了该实例的 collect_from_paths() 函数，且设置 b_execute_now=True，
                                此时若导入的成员中有类，且该类继承自某个父类，且在初始化时使用了 super(xx,self).__init__ 继承初始化函数，将出现
                                TypeError: super(type, obj): obj must be an instance or subtype of type 的错误
                            2. 在模块的 __init__.py 文件中使用 collect_from_paths()
                        为了避免情况 1，应该尽量避免设置 b_execute_now=True。
                            或者省略 super(xx,self).__init__ 中的参数改为 super().__init__
        """
        # 检查调用位置
        caller_frame = inspect.stack()[1]
        assert os.path.basename(caller_frame.filename) != "__init__.py", \
            f'calling Registry.collect_from_paths() in __init__.py is forbidden, file: {caller_frame.filename}.\n' \
            f'you can call it in other files, and then import the result of the call in __init__.py'

        #
        if not b_execute_now:
            self._path_to_collect.append(
                dict(path_ls=path_ls, path_ls_to_exclude=path_ls_to_exclude, b_execute_now=True))
            return

        #
        if path_ls is not None:
            temp = []
            for path in filter(lambda x: os.path.isdir(x), path_ls):
                temp.append(path)
                for root, dirs, _ in os.walk(path, topdown=False):
                    temp.extend([os.path.join(root, i) for i in dirs])
            if path_ls_to_exclude is not None:
                for path_ex in path_ls_to_exclude:
                    if not os.path.exists(path_ex):
                        continue
                    for i in reversed(range(len(temp))):
                        if os.path.samefile(os.path.commonpath([path_ex, temp[i]]), path_ex):
                            temp.pop(i)
            # 从深到浅导入，可以避免继承引起的 TypeError: super(type, obj) 类型错误
            path_ls = list(set(temp))
            path_ls.sort(reverse=True)

        temp = None
        for loader, module_name, is_pkg in pkgutil.walk_packages(path_ls):
            module = loader.find_module(module_name).load_module(module_name)
            if temp is None:
                for name, obj in inspect.getmembers(module):
                    if getattr(obj, "name", None) == Registry.name and getattr(obj, "uid", None) == self.uid:
                        temp = obj
                        break
        if temp is not None:
            self.database = temp.database

    # -------------------- 其他 --------------------- #

    @property
    def instances_of_Registry(self):
        return dict(Registry.__instances)


UNIFIED_REGISTRY = Registry(uid="UNIFIED_REGISTRY")



def escape_node(node, b_reversed=False, times=1):
    """
        对键进行转义/反转义
            转义是指在取值方式对应的几个特殊字符 :|@ 前面添加 \

        参数：
            node:               键
            b_reversed:         <boolean> 是否进行反转义
                                    默认为 False
            times:              <int> 转义的次数
                                    默认为 1
                                    当 b_reversed=True 时，times 设置为 -1 时候表示进行多次反转义，直至无法进行
    """
    if not isinstance(node, (str,)):
        node = f'{node}'
    assert isinstance(times, (int,))
    if times < 0:
        assert b_reversed, f'times can be set to -1 only when de-escaping'
        times = len(node)

    len_old = len(node)
    for _ in range(times):
        if b_reversed:
            node = node.replace("\:", ":").replace("\|", "|").replace("\@", "@")
        else:
            node = node.replace(":", "\:").replace("|", "\|").replace("@", "\@")
        if len(node) == len_old:
            break
        len_old = len(node)

    return node



from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import set_value_by_name, get_value_by_name


def eval_references(var, node_s, order, converter_for_ref=None, converter_for_res=None):
    """
        将 var 中的具有引用的值替换为计算结果

        参数：
            var:
            node_s:                 <dict> 引用节点，parse_references() 返回的结果
            order:                  <list of name> 计算顺序，cal_relation_between_references() 返回的结果
            converter_for_ref:      <callable> 对被引用节点施加何种处理
                                        形如 def(idx, v): ... 的函数，其中 idx 是被引用节点的名字，v是其值，
                                        返回的结果将替换掉被引用节点中原来的值。
                                        注意：
                                            - 处理后得到的结果将替换掉原引用节点的值。（重要所以说两次）
                                            - 当同一节点被多次引用时，仅会被处理、替换一次。
            converter_for_res:      <callable> 对计算结果施加何种处理
                                        形如 def(idx, v): ... 的函数，其中 idx 是节点的名字，v是计算结果
    """
    assert order is not None and set(order).issubset(set(node_s.keys()))
    assert converter_for_ref is None or callable(converter_for_ref)
    assert converter_for_res is None or callable(converter_for_res)

    processed_ref_nodes = set()

    for name in order:
        details = node_s[name]
        # 获取依赖值
        for k, idx in details["paras"].items():
            v_new = get_value_by_name(var=var, name=idx)
            if converter_for_ref is not None and idx not in processed_ref_nodes:
                v_new = converter_for_ref(idx, v_new)
                # 赋值
                set_value_by_name(var=var, name=idx, value=v_new, b_force=False)
                processed_ref_nodes.add(idx)
            details["paras"][k] = v_new
        # 计算
        res = eval(details["expression"], details["paras"])
        if converter_for_res is not None:
            res = converter_for_res(name, res)
        # 赋值
        set_value_by_name(var=var, name=name, value=res, b_force=False)

    return var



import re
from kevin_toolbox.computer_science.algorithm.for_nested_dict_list.name_handler import escape_node


def parse_name(name, b_de_escape_node=True):
    r"""
        解释名字 name 得到取值方式 method_ls 和取值时使用的键 node_ls

        参数：
            name:               <str> 名字
                                    由 "<变量名>" 加上多组 "<取值方式><键>" 组成。
                                    <变量名>
                                        在实际使用时，比如 get_value_by_name()等函数，一般忽略该部分。
                                    <取值方式>
                                        支持以下几种:
                                            "@"     表示使用 eval() 读取键
                                            ":"     表示使用 str() 读取键
                                            "|"     表示依次尝试 str() 和 eval() 两种方式
                                        示例:
                                            "@100"      表示读取 var[eval("100")]
                                            ":epoch"    表示读取 var["epoch"]
                                            "|1+1"    表示首先尝试读取 var["1+1"]，若不成功则尝试读取 var[eval("1+1")]
                                    <键>
                                        对于含有特殊字符 :|@ 的 node，应该先对 node 中的这些特殊字符使用 \ 进行转义
                                            比如，对于：
                                                var={"acc@epoch=1": 0.05, "acc@epoch=2": 0.06}
                                            在 var["acc@epoch=1"] 位置上的元素的名字可以是
                                                "var:acc\@epoch=1"
            b_de_escape_node:  <boolean> 是否尝试对取值的键名 node 进行反转义
                                    默认为 True
        返回：
            root_node, method_ls, node_ls
    """
    assert isinstance(name, (str,))

    temp = re.split(r'(?<!\\)([:@|])', name)
    root_node = temp.pop(0)
    method_ls, node_ls = temp[0::2], temp[1::2]
    if b_de_escape_node:
        node_ls = [escape_node(node=i, b_reversed=True, times=1) for i in node_ls]
    return root_node, method_ls, node_ls
